﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace WpfCosa
{
    class CosaApplication
    {
        private Boolean started = false;
        private Boolean running = false;
        private Int32 startTick = 0;
        private CellProcessor processor = null;
        private Dictionary<String,DataValue> valueMap = null;
        private Boolean readyToStop = false;

        public CosaApplication(CellProcessor cp, Int32 tick)
            : base()
        {
            processor = cp;
            startTick = tick;
            valueMap = new Dictionary<String,DataValue>();
        }

        public virtual String getTitle()
        {
            return "GenericApp";
        }

        protected virtual void init()
        {
        }

        public void start()
        {
            init();
            started = true;
        }

        protected virtual void stop()
        {
            started = false;
        }

        protected Int32 getStartTick()
        {
            return startTick;
        }

        protected Boolean doAction(Int32 currentTick)
        {
            doEffectors(currentTick);
            doSensors(currentTick);

            return readyToStop;
        }

        protected virtual void doEffectors(Int32 currentTick)
        {

        }

        protected virtual void doSensors(Int32 currentTick)
        {

        }

        public void run(Int32 currentTick)
        {
            if (running) return;

            running = true;
            Boolean readyToStop = doAction(currentTick);
            running = false;

            if (readyToStop)
            {
                stop();
            }
        }

        public Boolean isStarted()
        {
            return started;
        }

        protected void set(String valueName, Int32 i)
        {
            if (valueMap.ContainsKey(valueName))
            {
                DataValue value = valueMap[valueName];
                value.set(i);
            }
        }

        protected void set(String valueName, Double d)
        {
            if (valueMap.ContainsKey(valueName))
            {
                DataValue value = valueMap[valueName];
                value.set(d);
            }
        }

        protected void set(String valueName, String s)
        {
            if (valueMap.ContainsKey(valueName))
            {
                DataValue value = valueMap[valueName];
                value.set(s);
            }
        }

        protected void set(String valueName, Boolean b)
        {
            if (valueMap.ContainsKey(valueName))
            {
                DataValue value = valueMap[valueName];
                value.set(b);
            }
        }

        protected Int32 getInt(String valueName)
        {
            if (valueMap.ContainsKey(valueName))
            {
                DataValue value = valueMap[valueName];
                return value.getInt();
            }
            else
            {
                return 0;
            }
        }

        protected Double getDouble(String valueName)
        {
            if (valueMap.ContainsKey(valueName))
            {
                DataValue value = valueMap[valueName];
                return value.getDouble();
            }
            else
            {
                return 0;
            }
        }

        protected String getString(String valueName)
        {
            if (valueMap.ContainsKey(valueName))
            {
                DataValue value = valueMap[valueName];
                return value.getString();
             }
            else
            {
                return "";
            }
        }

        protected Boolean getBoolean(String valueName)
        {
            if (valueMap.ContainsKey(valueName))
            {
                DataValue value = valueMap[valueName];
                return value.getBoolean();
            }
            else
            {
                return false;
            }
        }

        protected void initValue(String initText)
        {
            DataValue value = DataValue.newInstance(initText);
            String valueName = value.getName();

            // validate value name does not already exist
            if (valueMap.ContainsKey(valueName)) return;

            // add to value map
            valueMap.Add(valueName, value);
        }

        protected void runCommand(String commandText)
        {
            Parser parser = new Parser(commandText);
            String command = parser.nextToken();

            // if no command text, skip
            if (command.Length == 0) return;

            // process the command
            String textParam = "";
            String varParam = "";
            String newValue = "";

            String param1 = "";
            String param2 = "";
            Double result = 0.0;

            switch (command)
            {
                case "outputLine":
                    textParam = Parser.stripQuotes(parser.nextToken());
                    Console.WriteLine(replaceVariables(textParam));
                    break;
                case "output":
                    textParam = Parser.stripQuotes(parser.nextToken());
                    Console.WriteLine(replaceVariables(textParam));
                    break;
                case "increment":
                    varParam = parser.nextToken();
                    Int32 value = getInt(varParam);
                    value++;

                    set(varParam, value);
                    break;
                case "set":
                    varParam = parser.nextToken();
                    newValue = replaceVariables(parser.nextToken());

                    // set the value
                    setValue(varParam, newValue);
                    break;
                case "add":
                    varParam = parser.nextToken();
                    param1 = replaceVariables(parser.nextToken());
                    param2 = replaceVariables(parser.nextToken());

                    // convert to double and perform math
                    result = Double.Parse(param1) + Double.Parse(param2);

                    // set the value
                    setValue(varParam, result);
                    break;
                case "subtract":
                    varParam = parser.nextToken();
                    param1 = replaceVariables(parser.nextToken());
                    param2 = replaceVariables(parser.nextToken());

                    // convert to double and perform math
                    result = Double.Parse(param1) - Double.Parse(param2);

                    // set the value
                    setValue(varParam, result);
                    break;
                case "multiply":
                    varParam = parser.nextToken();
                    param1 = replaceVariables(parser.nextToken());
                    param2 = replaceVariables(parser.nextToken());

                    // convert to double and perform math
                    result = Double.Parse(param1) * Double.Parse(param2);

                    // set the value
                    setValue(varParam, result);
                    break;
                case "divide":
                    varParam = parser.nextToken();
                    param1 = replaceVariables(parser.nextToken());
                    param2 = replaceVariables(parser.nextToken());

                    // convert to double and perform math
                    result = Double.Parse(param1) / Double.Parse(param2);

                    // set the value
                    setValue(varParam, result);
                    break;
                case "stop":
                    readyToStop = true;
                    break;
                case "lessThan":
                    param1 = replaceVariables(parser.nextToken());
                    param2 = replaceVariables(parser.nextToken());

                    if (Double.Parse(param1) < Double.Parse(param2))
                    {
                        runCommand(parser.getContent());
                    }
                    break;
                case "equals":
                    param1 = replaceVariables(parser.nextToken());
                    param2 = replaceVariables(parser.nextToken());

                    if (Double.Parse(param1) == Double.Parse(param2))
                    {
                        runCommand(parser.getContent());
                    }
                    break;
            }
        }

        private String replaceVariables(String text)
        {
            foreach (KeyValuePair<String,DataValue> kvp in valueMap)
            {
                String valueName = kvp.Key.ToString();
                DataValue value = kvp.Value;
                text = text.Replace("$[" + valueName + "]", value.getString());
            }

            return text;
        }

        private void setValue(String resultParam, String s)
        {
            // determine the result value type
            DataValue value = valueMap[resultParam];
            String valueType = value.getValueType();

            // convert to the appropriate type and set
            if (valueType == "int")
            {
                Int32 cValue = 0;
                try { cValue = Int32.Parse(s); }
                catch { }
                set(resultParam, cValue);
            }
            else if (valueType == "double")
            {
                Double cValue = 0;
                try { cValue = Double.Parse(s); }
                catch { }
                set(resultParam, cValue);
            }
            else if (valueType == "string")
            {
                String cValue = Parser.stripQuotes(s);
                set(resultParam, cValue);
            }
            else if (valueType == "boolean")
            {
                Boolean cValue = false;
                try { cValue = Boolean.Parse(s); }
                catch { }
                set(resultParam, cValue);
            }
        }

        // sets a numeric value to the specified variable
        private void setValue(String resultParam, Double d)
        {
            // determine the result value type
            DataValue value = valueMap[resultParam];
            String valueType = value.getValueType();
            
            // convert to the appropriate type and set
            Int32 iValue = 0;
            Double dValue = 0.0;

            switch (valueType) {
                case "int":
                    try { iValue = Convert.ToInt32(d); }
                    catch { }
                    set(resultParam, iValue);

                    break;
                case "double":
                    try { dValue = Convert.ToDouble(d); }
                    catch { }
                    set(resultParam, dValue);

                    break;
            }
        }

        protected Boolean isReadyToStop()
        {
            return readyToStop;
        }
    }
}
